"""Module for interface to KLF 200."""
import json
import asyncio
import aiohttp
import async_timeout

from .exception import PyVLXException, InvalidToken


class Interface:
    """Interface to KLF 200."""

    def __init__(self, config, timeout=10):
        """Initialize interface class."""
        self.config = config
        self.token = None
        self.timeout = timeout

    async def api_call(self, verb, action, params=None, add_authorization_token=True, retry=False):
        """Send api call."""
        if add_authorization_token and not self.token:
            await self.refresh_token()

        try:
            return await self._api_call_impl(verb, action, params, add_authorization_token)
        except InvalidToken:
            if not retry and add_authorization_token:
                await self.refresh_token()
                # Recursive call of api_call
                return await self.api_call(verb, action, params, add_authorization_token, True)
            raise

    async def _api_call_impl(self, verb, action, params=None, add_authorization_token=True):
        url = self.create_api_url(self.config.host, verb)
        body = self.create_body(action, params)
        headers = self.create_headers(add_authorization_token, self.token)
        return await self._do_http_request(url, body, headers)

    async def _do_http_request(self, url, body, headers):
        try:
            return await self._do_http_request_impl(url, body, headers)
        except asyncio.TimeoutError:
            raise PyVLXException("Request timeout when talking to VELUX API")
        except aiohttp.ClientError:
            raise PyVLXException("HTTP error when talking to VELUX API")
        except OSError:
            raise PyVLXException("OS error when talking to VELUX API")

    async def _do_http_request_impl(self, url, body, headers):
        print(url, body, headers)
        async with aiohttp.ClientSession() as session:
            with async_timeout.timeout(self.timeout):
                async with session.post(url, data=json.dumps(body), headers=headers) as response:
                    response = await response.text()
                    response = self.fix_response(response)
                    print(response)
                    json_response = json.loads(response)
                    self.evaluate_response(json_response)
                    # print(json.dumps(json_response, indent=4, sort_keys=True))
                    return json_response

    async def refresh_token(self):
        """Refresh API token from KLF 200."""
        json_response = await self.api_call('auth', 'login', {'password': self.config.password}, add_authorization_token=False)
        if 'token' not in json_response:
            raise PyVLXException('no element token found in response: {0}'.format(json.dumps(json_response)))
        self.token = json_response['token']

    async def disconnect(self):
        """Disconnect from KLF 200."""
        await self.api_call('auth', 'logout', {}, add_authorization_token=True)
        self.token = None

    @staticmethod
    def create_api_url(host, verb):
        """Return full rest url."""
        return 'http://{0}/api/v1/{1}'.format(host, verb)

    @staticmethod
    def create_headers(add_authorization_token, token=None):
        """Create http header for rest request."""
        headers = {}
        headers['Content-Type'] = 'application/json'
        if add_authorization_token:
            headers['Authorization'] = 'Bearer ' + token
        return headers

    @staticmethod
    def create_body(action, params):
        """Create http body for rest request."""
        body = {}
        body['action'] = action
        if params is not None:
            body['params'] = params
        return body

    @staticmethod
    def evaluate_response(json_response):
        """Evaluate rest response."""
        if 'errors' in json_response and json_response['errors']:
            Interface.evaluate_errors(json_response)
        elif 'result' not in json_response:
            raise PyVLXException('no element result  found in response: {0}'.format(json.dumps(json_response)))
        elif not json_response['result']:
            raise PyVLXException('Request failed {0}'.format(json.dumps(json_response)))

    @staticmethod
    def evaluate_errors(json_response):
        """Evaluate rest errors."""
        if 'errors' not in json_response or \
           not isinstance(json_response['errors'], list) or \
           not json_response['errors'] or \
           not isinstance(json_response['errors'][0], int):
            raise PyVLXException('Could not evaluate errors {0}'.format(json.dumps(json_response)))

        # unclear if response may contain more errors than one. Taking the first.
        first_error = json_response['errors'][0]

        if first_error in [402, 403, 405, 406]:
            raise InvalidToken(first_error)

        raise PyVLXException('Unknown error code {0}'.format(first_error))

    @staticmethod
    def fix_response(response):
        """Fix broken rest reponses."""
        # WTF: For whatever reason, the KLF 200 sometimes puts an ')]}',' in front of the response ...
        index = response.find('{')
        if index > 0:
            return response[index:]
        return response



"""Unit tests for FrameGetLimitationStatus."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames.frame_get_limitation import FrameGetLimitationStatus
from pyvlx.const import LimitationType


class TestFrameGetLimitationStatus(unittest.TestCase):
    """Test class for FrameGetLimitationStatus."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetLimitationStatus bytes."""
        frame = FrameGetLimitationStatus(node_ids=[1], session_id=1, limitation_type=LimitationType.MIN_LIMITATION)
        self.assertEqual(bytes(frame), b'\x00\x1c\x03\x12\x00\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                                       b'\x00\x00\x0c')

        frame = FrameGetLimitationStatus(node_ids=[1, 2], session_id=2, limitation_type=LimitationType.MAX_LIMITATION)
        self.assertEqual(bytes(frame), b'\x00\x1c\x03\x12\x00\x02\x02\x01\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                                       b'\x00\x01\x0f')

    def test_frame_from_raw(self):
        """Test parse FrameGetLimitationStatus from raw."""
        frame = frame_from_raw(b'\x00\x1c\x03\x12\x00\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                               b'\x0c')
        self.assertTrue(isinstance(frame, FrameGetLimitationStatus))

    def test_str(self):
        """Test string representation of FrameGetLimitationStatus."""
        frame = FrameGetLimitationStatus(node_ids=[1], session_id=1, limitation_type=LimitationType.MIN_LIMITATION)
        self.assertEqual(str(frame), '<FrameGetLimitationStatus node_ids="[1]" session_id="1" originator="Originator.USER" />')



"""Test for slip helper functions."""
import unittest
from typing import Any

from pyvlx.slip import decode, encode, get_next_slip, is_slip, slip_pack


class TestSlip(unittest.TestCase):
    """Test class for slip helper functions."""

    # pylint: disable=invalid-name

    def encode_decode(self, decoded: Any, encoded: Any) -> None:
        """Decode encoded, encode decoded and test results."""
        self.assertEqual(decode(encoded), decoded)
        self.assertEqual(encode(decoded), encoded)

    def test_simple(self) -> None:
        """Test decode and encode."""
        self.encode_decode(
            b"\xc0\xc0\xdb\xdb\xc0\xdb\xc0",
            b"\xdb\xdc\xdb\xdc\xdb\xdd\xdb\xdd\xdb\xdc\xdb\xdd\xdb\xdc",
        )

    def test_empty(self) -> None:
        """Test decode and  encode of empty string."""
        self.encode_decode(b"", b"")

    def test_real_live_example(self) -> None:
        """Test decode of real live example."""
        self.encode_decode(
            b"\x00\x10\x03\x02\x00\x01\x01\x00\x00\xc8\x00\x02\x1e\x06\x00\x03\x00\xc0",
            b"\x00\x10\x03\x02\x00\x01\x01\x00\x00\xc8\x00\x02\x1e\x06\x00\x03\x00\xdb\xdc",
        )

    def test_slip(self) -> None:
        """Test is_slip function."""
        self.assertTrue(is_slip(b"\xc0abc\xc0xx"))
        self.assertFalse(is_slip(b"zz\xc0abc\xc0"))
        self.assertFalse(is_slip(b"\xc0abc"))
        self.assertFalse(is_slip(b"\xc0"))

    def test_slip_pack(self) -> None:
        """Test slip_pack function."""
        self.assertEqual(
            slip_pack(b"\xc0\xc0\xdb\xdb\xc0\xdb\xc0"),
            b"\xc0\xdb\xdc\xdb\xdc\xdb\xdd\xdb\xdd\xdb\xdc\xdb\xdd\xdb\xdc\xc0",
        )

    def test_get_next_slip(self) -> None:
        """Test get_next_slip function."""
        self.assertEqual(
            get_next_slip(
                b"\xc0\xdb\xdc\xdb\xdc\xdb\xdd\xdb\xdd\xdb\xdc\xdb\xdd\xdb\xdc\xc0fnord"
            ),
            (b"\xc0\xc0\xdb\xdb\xc0\xdb\xc0", b"fnord"),
        )

    def test_get_next_slip_no_complete_slip(self) -> None:
        """Test get_next_slip function with no complete slip frame beeing in queue."""
        self.assertEqual(get_next_slip(b"zz\xc0abc\xc0"), (None, b"zz\xc0abc\xc0"))

    def test_get_next_slip_empty_queue(self) -> None:
        """Test get_next_slip function with empty queue."""
        self.assertEqual(get_next_slip(b""), (None, b""))



"""Unit tests for FramePasswordEnterConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FramePasswordEnterConfirmation, PasswordEnterConfirmationStatus)


class TestFramePasswordEnterConfirmation(unittest.TestCase):
    """Test class for FramePasswordEnterConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FramePasswordEnterConfirmation."""
        frame = FramePasswordEnterConfirmation()
        self.assertEqual(bytes(frame), b"\x00\x040\x01\x005")

    def test_bytes_error(self):
        """Test failed FramePasswordEnterConfirmation."""
        frame = FramePasswordEnterConfirmation(
            status=PasswordEnterConfirmationStatus.FAILED
        )
        self.assertEqual(bytes(frame), b"\x00\x040\x01\x014")

    def test_frame_from_raw(self):
        """Test parse FramePasswordEnterConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x040\x01\x014")
        self.assertTrue(isinstance(frame, FramePasswordEnterConfirmation))
        self.assertEqual(frame.status, PasswordEnterConfirmationStatus.FAILED)

    def test_str(self):
        """Test string representation of FramePasswordEnterConfirmation."""
        frame = FramePasswordEnterConfirmation()
        self.assertEqual(
            str(frame),
            '<FramePasswordEnterConfirmation status="PasswordEnterConfirmationStatus.SUCCESSFUL"/>',
        )



"""Test for slip helper functions."""
import unittest

from pyvlx.api.session_id import get_new_session_id, set_session_id


class SessionIdSlip(unittest.TestCase):
    """Test class for slip helper functions."""

    # pylint: disable=invalid-name

    def test_session_id(self):
        """Decode encoded, encode decoded and test results."""
        set_session_id(0)  # Reset session id
        self.assertEqual(get_new_session_id(), 1)
        self.assertEqual(get_new_session_id(), 2)
        self.assertEqual(get_new_session_id(), 3)
        set_session_id(65535 - 1)
        self.assertEqual(get_new_session_id(), 65535)
        self.assertEqual(get_new_session_id(), 1)
